﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Threading;
using SlimDX;
using SlimDX.DirectSound;

namespace FDK
{
	public class CSoundDeviceDirectSound : ISoundDevice
	{
		// プロパティ

		public ESoundDeviceType e出力デバイス
		{
			get;
			protected set;
		}
		public long n実出力遅延ms
		{
			get;
			protected set;
		}
		public static readonly BufferFlags DefaultFlags = BufferFlags.Defer | BufferFlags.GetCurrentPosition2 | BufferFlags.GlobalFocus | BufferFlags.ControlVolume | BufferFlags.ControlPan | BufferFlags.ControlFrequency;

		// CSoundTimer 用に公開しているプロパティ

		public long n経過時間ms
		{
			get
			{
				int n現在位置 = this.sd経過時間計測用サウンドバッファ.DirectSoundBuffer.CurrentPlayPosition;

				int nループ回数 = 0;
				lock( this.objループ回数同期用 )
				{
					nループ回数 = this.nループ回数;
				}

				long n経過時間ms = (long) ( ( nループ回数 * n単位繰り上げ間隔 * 1000.0 ) + ( n現在位置 * 1000.0 / ( 44100.0 * 2 * 2 ) ) );

				while( n経過時間ms < this.n前回の経過時間ms )	// 前回より小さければループ回数が足りない可能性があるので
					n経過時間ms += n単位繰り上げ間隔;			// 足りるまで足してやる。

				this.n前回の経過時間ms = n経過時間ms;

				return n経過時間ms;
			}
		}
		public long n経過時間を更新したシステム時刻ms
		{
			get { throw new NotImplementedException(); }
		}
		public CTimer tmシステムタイマ
		{
			get;
			protected set;
		}


		// メソッド

		public CSoundDeviceDirectSound( IntPtr hWnd, long n遅延時間ms )
		{
			this.e出力デバイス = ESoundDeviceType.Unknown;
			this.n実出力遅延ms = n遅延時間ms;
			this.tmシステムタイマ = new CTimer( CTimer.E種別.MultiMedia );

			#region [ DirectSound デバイスを作成する。]
			//-----------------
			this.DirectSound = new DirectSound();	// 失敗したら例外をそのまま発出。

			// デバイスの協調レベルを設定する。

			bool priority = true;
			try
			{
				this.DirectSound.SetCooperativeLevel( hWnd, CooperativeLevel.Priority );
			}
			catch( DirectSoundException )
			{
				this.DirectSound.SetCooperativeLevel( hWnd, CooperativeLevel.Normal );	// これでも失敗したら例外をそのまま発出。
				priority = false;
			}

			// デバイス作成完了。

			this.e出力デバイス = ESoundDeviceType.DirectSound;
			//-----------------
			#endregion

			#region [ 経過時間計測用サウンドバッファと単位繰り上げ用スレッドを作成し、ループ再生を開始する。]
			//-----------------
			
			// 単位繰り上げ間隔[秒]の長さを持つ無音のサウンドを作成。

			uint nデータサイズbyte = n単位繰り上げ間隔 * 44100 * 2 * 2;
			var ms = new MemoryStream();
			var bw = new BinaryWriter( ms );
			bw.Write( (uint) 0x46464952 );						// 'RIFF'
			bw.Write( (uint) ( 44 + nデータサイズbyte - 8 ) );	// ファイルサイズ - 8
			bw.Write( (uint) 0x45564157 );						// 'WAVE'
			bw.Write( (uint) 0x20746d66 );						// 'fmt '
			bw.Write( (uint) 16 );								// バイト数
			bw.Write( (ushort) 1 );								// フォーマットID(リニアPCM)
			bw.Write( (ushort) 2 );								// チャンネル数
			bw.Write( (uint) 44100 );							// サンプリング周波数
			bw.Write( (uint) ( 44100 * 2 * 2 ) );				// bytes/sec
			bw.Write( (ushort) ( 2 * 2 ) );						// blockサイズ
			bw.Write( (ushort) 16 );							// bit/sample
			bw.Write( (uint) 0x61746164 );						// 'data'
			bw.Write( (uint) nデータサイズbyte);				// データ長
			for( int i = 0; i < nデータサイズbyte / sizeof(long); i++ )	// PCMデータ
				bw.Write( (long) 0 );
			var byArrWaveFleImage = ms.ToArray();
			bw.Close();
			ms = null;
			bw = null;
			this.sd経過時間計測用サウンドバッファ = this.tサウンドを作成する( byArrWaveFleImage, CSoundDeviceDirectSound.DefaultFlags | BufferFlags.ControlPositionNotify );	// 位置通知 ON
			CSound.listインスタンス.Remove( this.sd経過時間計測用サウンドバッファ );	// 特殊用途なのでインスタンスリストからは除外する。

			// 開始点（位置0）にきたらイベントを通知する用設定。
			
			this.sd経過時間計測用サウンドバッファ.DirectSoundBuffer.SetNotificationPositions( new NotificationPosition[ 1 ] { new NotificationPosition() { Offset = 0, Event = this.autoResetEvent } } );

			// イベントを受けるスレッドを作成し、実行開始。
			
			this.th経過時間測定用スレッド = new Thread( new ThreadStart( this.tループ回数をカウントする ) );
			this.th経過時間測定用スレッド.Start();			// 経過時間測定スレッド開始。

			// サウンドのループ再生開始。
			
			this.sd経過時間計測用サウンドバッファ.DirectSoundBuffer.Play( 0, PlayFlags.Looping );
			//-----------------
			#endregion

			Trace.TraceInformation( "DirectSound を初期化しました。({0})", ( priority ) ? "Priority" : "Normal" );
		}

		public CSound tサウンドを作成する( string strファイル名 )
		{
			var sound = new CSound();
			sound.tDirectSoundサウンドを作成する( strファイル名, this.DirectSound );
			return sound;
		}
		public CSound tサウンドを作成する( byte[] byArrWAVファイルイメージ )
		{
			var sound = new CSound();
			sound.tDirectSoundサウンドを作成する( byArrWAVファイルイメージ, this.DirectSound );
			return sound;
		}
		public CSound tサウンドを作成する( byte[] byArrWAVファイルイメージ, BufferFlags flags )
		{
			var sound = new CSound();
			sound.tDirectSoundサウンドを作成する( byArrWAVファイルイメージ, this.DirectSound, flags );
			return sound;
		}

		// 既存のインスタンス（生成直後 or Dispose済み）に対してサウンドを生成する。
		public void tサウンドを作成する( string strファイル名, ref CSound sound )
		{
			sound.tDirectSoundサウンドを作成する( strファイル名, this.DirectSound );
		}
		public void tサウンドを作成する( byte[] byArrWAVファイルイメージ, ref CSound sound )
		{
			sound.tDirectSoundサウンドを作成する( byArrWAVファイルイメージ, this.DirectSound );
		}
		public void tサウンドを作成する( byte[] byArrWAVファイルイメージ, BufferFlags flags, ref CSound sound )
		{
			sound.tDirectSoundサウンドを作成する( byArrWAVファイルイメージ, this.DirectSound, flags );
		}

		#region [ Dispose-Finallizeパターン実装 ]
		//-----------------
		public void Dispose()
		{
			this.Dispose( true );
			GC.SuppressFinalize( this );
		}
		protected void Dispose( bool bManagedDispose )
		{
			if( bManagedDispose )
			{
				#region [ 経緯時間計測用サウンドバッファを解放。]
				//-----------------
				this.sd経過時間計測用サウンドバッファ.tサウンドを停止する();
				C共通.tDisposeする( ref this.sd経過時間計測用サウンドバッファ );
				//-----------------
				#endregion
				#region [ 単位繰り上げ用スレッド停止。]
				//-----------------
				if( this.th経過時間測定用スレッド != null )
				{
					this.th経過時間測定用スレッド.Abort();
					this.th経過時間測定用スレッド = null;
				
				}
				//-----------------
				#endregion

				C共通.tDisposeする( ref this.DirectSound );
				C共通.tDisposeする( this.tmシステムタイマ );
			}
		}
		~CSoundDeviceDirectSound()
		{
			this.Dispose( false );
		}
		//-----------------
		#endregion

		protected DirectSound DirectSound = null;
		protected CSound sd経過時間計測用サウンドバッファ = null;
		protected Thread th経過時間測定用スレッド = null;
		protected AutoResetEvent autoResetEvent = new AutoResetEvent( false );
		protected const uint n単位繰り上げ間隔 = 1;	// [秒]
		protected int nループ回数 = 0;
		protected object objループ回数同期用 = new object();
		
		protected void tループ回数をカウントする()
		{
			while( true )
			{
				// 経過時間計測用サウンドバッファの再生カーソルがループして先頭に来るのを待つ。

				this.autoResetEvent.WaitOne();


				// ループ回数をカウントアップする。

				lock( this.objループ回数同期用 )
				{
					this.nループ回数++;
				}
			}
		}

		private long n前回の経過時間ms = -1;
	}
}
